jQuery.sap.declare("sas.stress.wkbench.widget.RiskTable");

jQuery.sap.require("sas.stress.wkbench.widget.RiskTable");
jQuery.sap.require("sas.hc.ui.table.Table");
jQuery.sap.require("sas.hc.ui.table.TableRenderer");


/**
 * Extension of sas.hc.ui.table.Table that will preserve table selections even after filtering or sorting
 *
 * @name sas.stress.wkbench.widget.RiskTable
 */
sas.hc.ui.table.Table.extend("sas.stress.wkbench.widget.RiskTable", {
    metadata: {
        properties: {
            /**
             * Flag used to remember the table row selection through out user session and reapply on table data change
             */
            persistSelectionsThroughSession: {type: "boolean", defaultValue: false},
            /**
             * Unique id that will be used to identify the table for which persisted rows will be selected
             */
            persistenceId: {type: "string"},
            /**
             * Array of attributes that can uniquely identify the row object. Default is to use "id".
             */
            rowUniqueIdentifiers: {type: "string[]", defaultValue: ["id"]}
        },
        events: {
            /**
             * fired when the preserved selection context paths array values change
             */
            selectionUpdated: {
                parameters: {
                    selectionPaths: {type : "string[]"}
                }
            }
        }
    },
    renderer: "sas.hc.ui.table.TableRenderer"
});

sas.stress.wkbench.widget.RiskTable.prototype.onBeforeRendering = function(){
    var that = this;
    if(this.getPersistSelectionsThroughSession()){
        if(_.isEmpty(this.getPersistenceId()) || _.isEmpty(this.getRowUniqueIdentifiers())){
            console.error("Missing attributes: Either persistenceId or rowUniqueIdentifiers or both are missing. Therefore 'persistSelectionsThroughSession' attribute will be ignored.");
            this.setPersistSelectionsThroughSession(false);
        }
    }

    sas.hc.ui.table.Table.prototype.onBeforeRendering.apply(this, arguments);
};

sas.stress.wkbench.widget.RiskTable.prototype.onAfterRendering = function(oEvent) {
    var that = this;
    var selectAllCheckBox = this.getSelectAllControl();
    selectAllCheckBox.attachSelect(function(e){
        //If un checking the select all check box, then clear all preserved selections
        var unselected = !e.getParameter("selected");
        if(unselected){
            //If it is a filtered list, then only clear the selection paths for the current filtered list of rows.
            //Otherwise clear all the selection paths
            if(that.isFiltered()){
                var currentContexts = that.getBinding("rows").getContexts(0);
                var currentPaths = _.map(currentContexts, function(context){
                    return context.getPath();
                });
                that.removeSelectedContextPaths(currentPaths);
            }else{
                that.clearSelectedContextPaths();
            }
        }
    });
    sas.hc.ui.table.Table.prototype.onAfterRendering.apply(this, arguments);
};

// Internal private methods tapping - START

//Getting hold of the _updateBindingContexts which is called when the table rows update is complete.
//Using other change handlers like _onBindingChange is not useful since by then the rows update is not complete yet.
//This is where the preserved selections will be applied
sas.stress.wkbench.widget.RiskTable.prototype._updateBindingContexts = function(bSuppressUpdate, iRowCount, sReason) {
    sas.hc.ui.table.Table.prototype._updateBindingContexts.apply(this, arguments);

    //If the reason we are here is due to a binding change (ex: setData call), and if persistSelectionsThroughSession flag is on, then update the remembered context paths from the persisted rows in session storage
    if(sReason === "change" && this.getPersistSelectionsThroughSession()){
        this.clearSelection(); //Clear any existing selection on binding change so that we can put the right selections back
        this.clearSelectedContextPaths(true);//Is helpful when same table instance is used but the persisted rows are cleared from application
        this.updateSelectedContextPaths(); //binding has changed. keep the selectedContextPaths list up to date
        this.handleBindingContextsUpdated();
        // this.applyPersistedRows();
    }else if(sReason !== "change"){ //If we are here due to sorting or filtering then reapply the remembered selections
        this.handleBindingContextsUpdated();
    }
};

//Hooking into the selectionModel's selectionChange handler to populate the selected contexts and indices properly before any rowSelectionChange handlers are called
//Table's _onSelectionChanged function is the one which fires the rowSelectionChange event. Therefore need to call the handleSelectionChange here before the rowSelectionChange is fired
sas.stress.wkbench.widget.RiskTable.prototype._onSelectionChanged = function(oEvent) {
    var changedIndices = oEvent.getParameter("rowIndices");
    var oldIndex = oEvent.getParameter("oldIndex");
    var leadIndex = oEvent.getParameter("leadIndex");
    var clearSelectionCase = oldIndex !== -1 && leadIndex === -1;
    var selectedIndices = this.getSelectedIndices();
    var addSelectionIndices = [], removeSelectionIndices = [], changedIndex;
    for(var i=0; i<changedIndices.length; i++){
        changedIndex = changedIndices[i];
        if(_.contains(selectedIndices, changedIndex)){
            addSelectionIndices.push(changedIndex);
        }else{
            //Not adding the remove selection indices here when "clearSelection" is called. This will cover the setSelectionInterval case when we might be checking some and unchecking some
            //Rather the removing part is done through "removeSelectionInterval" which is called with user deselect event or with manual call to removeSelectionInterval.
            //This way we can avoid clearing our remembered selections in case of table's "clearSelection" call which is called either explicitly or when for example table is filtered or sorted.
            //When using RiskTable, consumer needs to call "clearSelectedContextPaths" along with clearSelection when they truly mean to forget the remembered selections.
            if(!clearSelectionCase){
                removeSelectionIndices.push(changedIndex);
            }
        }
    }
    this.handleSelectionChange(addSelectionIndices, removeSelectionIndices);
    sas.hc.ui.table.Table.prototype._onSelectionChanged.apply(this, arguments);
};

//Adding hook to clear context paths selection when this selection removal method is called.
//It's because this is a non-user interaction case and here the params passed to _onSelectionChanged event will result in the calculated rowIndex to be -1
sas.stress.wkbench.widget.RiskTable.prototype.removeSelectionInterval = function (iFromIndex, iToIndex) {
    var pathsToClear = [];
    //remove the implicitly removed (no user interaction) selection detail from the selected context paths
    for(var i = iFromIndex; i <= iToIndex; i++){
        var context = this.getContextByIndex(i);
        if(context.getPath()){
            pathsToClear.push(context.getPath());
        }
    }
    if(pathsToClear.length>0){
        this.removeSelectedContextPaths(pathsToClear);
    }

    sas.hc.ui.table.Table.prototype.removeSelectionInterval.apply(this, arguments);
};

//When it is a single select table the action taken is to deselect a row, then handle it using removeSelectionInterval instead of the default action of calling clearSelection
//RiskTable avoids actions on clearSelection because that is called during table filter / sorting as well. The core of RiskTable is to keep the selection on table filter or sorting.
//For multi select table, by default it calls removeSelectionInterval for deselection of rows
sas.stress.wkbench.widget.RiskTable.prototype._onRowSelect = function(iRowIndex, bShift, bCtrl){
    if (this.getSelectionMode() === sas.hc.ui.table.SelectionMode.Single && this.isIndexSelected(iRowIndex)) {
        this.removeSelectionInterval(iRowIndex, iRowIndex);
    }else{
        sas.hc.ui.table.Table.prototype._onRowSelect.apply(this, arguments);
    }
};

// Internal private methods tapping - END

/**
 * Consumers who wish to get hold of *all* of the selected rows (visible or not), should call this method to get all the selected rows context paths.
 * Calling getSelectedIndices on the table will only return *visible* selected rows indices
 *
 * @returns {array of binding context paths for the selected rows}
 */
sas.stress.wkbench.widget.RiskTable.prototype.getSelectedContextPaths = function(){
    var selectionContextsCustomData = this.findSelectionContextsCustomData();
    var selectedPaths = selectionContextsCustomData ? selectionContextsCustomData.getValue() : [];
    return selectedPaths;
};

//Clear the preserved selection cache
sas.stress.wkbench.widget.RiskTable.prototype.clearSelectedContextPaths = function(keepPersistedRows){
    var selectionContextsCustomData = this.findSelectionContextsCustomData();
    if(selectionContextsCustomData){
        selectionContextsCustomData.setValue([]);
    }
    if(this.getPersistSelectionsThroughSession() && !keepPersistedRows) {
        this.clearPersistedRows();
    }
    this.fireEvent("selectionUpdated",{"selectionPaths": []});
};

//Remove the given paths from the selected list
sas.stress.wkbench.widget.RiskTable.prototype.removeSelectedContextPaths = function(pathsArray){
    var selectionContextsCustomData = this.findSelectionContextsCustomData();
    var selectedPaths = selectionContextsCustomData ? selectionContextsCustomData.getValue() : [];
    _.each(pathsArray, function(path){
        selectedPaths = _.without(selectedPaths, path);
    });
    selectionContextsCustomData.setValue(selectedPaths);

    if(this.getPersistSelectionsThroughSession()){
        this.removeAndPersistSelectedRows(pathsArray);
    }

    this.fireEvent("selectionUpdated",{"selectionPaths": selectedPaths});
};

//Returns true if the path is one of the selected paths
sas.stress.wkbench.widget.RiskTable.prototype.isSelectedContextPath = function(path) {
    return $.inArray(path, this.getSelectedContextPaths()) !== -1;
};

//This is called when the table data is updated. Now apply the preserved selections.
sas.stress.wkbench.widget.RiskTable.prototype.handleBindingContextsUpdated = function(){
    var selectionContextPaths = this.getSelectedContextPaths();
    if(selectionContextPaths.length===0){ //Exit early if nothing to select anyway
        return;
    }

    var contexts = this.getBinding("rows").getContexts(0);
    var selectedIndices = this.getSelectedIndices();
    var i;
    for(i=0; i<contexts.length; i++){
        var rowBindingContextPath = contexts[i].getPath();
        if(_.contains(selectionContextPaths, rowBindingContextPath) && (!_.contains(selectedIndices, i))){
            this.addSelectionInterval(i, i);
        }
    }
};

//This is where the selections are preserved.
sas.stress.wkbench.widget.RiskTable.prototype.handleSelectionChange = function(addSelectionIndices, removeSelectionIndices){
    //For single selection table, clear any previous selection so that it can preserve only the last selection below.
    if(addSelectionIndices.length > 0 && this.getSelectionMode()===sas.hc.ui.table.SelectionMode.Single){
        this.clearSelectedContextPaths();
    }

    var selectionContextsCustomData = this.findOrAddSelectionContextsCustomData();
    var selectionContextPaths = this.getSelectedContextPaths();
    var contextPathsAdded=[], contextPathsRemoved=[];
    var addSelectionIndex, removeSelectionIndex, context, contextPath;
    for(var i=0; i<addSelectionIndices.length; i++){
        addSelectionIndex = addSelectionIndices[i];
        context = this.getContextByIndex(addSelectionIndex);
        //checking for context not undefined is useful when table has more selected rows, but the number of rows shown is less due to filter(especially homegrown filter like MIP's advanced filter)
        if(!context){
            continue;
        }
        contextPath = context.getPath();
        if(!_.contains(selectionContextPaths, contextPath)){
            contextPathsAdded.push(contextPath);
            selectionContextPaths.push(context.getPath());
        }
    }
    for(var i=0; i<removeSelectionIndices.length; i++){
        removeSelectionIndex = removeSelectionIndices[i];
        context = this.getContextByIndex(removeSelectionIndex);
        if(!!context){
            contextPath = context.getPath();
            contextPathsRemoved.push(contextPath);
            selectionContextPaths = _.without(selectionContextPaths, contextPath);
        }
    }
    selectionContextsCustomData.setValue(selectionContextPaths);
    this.fireEvent("selectionUpdated",{"selectionPaths": selectionContextPaths});

    if(this.getPersistSelectionsThroughSession()){
        var persistedRowsArray = this.getPersistedRowsArray();
        this.addPersistedRows(contextPathsAdded, persistedRowsArray);
        this.removePersistedRows(contextPathsRemoved, persistedRowsArray);
        this.updatePersistedRowsInStorage(persistedRowsArray);
    }
};

sas.stress.wkbench.widget.RiskTable.prototype.findSelectionContextsCustomData = function(){
    return _.find(this.getCustomData(), function(cData){
        return cData.getKey() === "selectedContextPaths";
    });
};

sas.stress.wkbench.widget.RiskTable.prototype.findOrAddSelectionContextsCustomData = function(){
    var selectionContextsCustomData = this.findSelectionContextsCustomData();
    if(!selectionContextsCustomData){
        selectionContextsCustomData = new sap.ui.core.CustomData({
            key: "selectedContextPaths",
            value: []
        });
        this.addCustomData(selectionContextsCustomData);
    }
    return selectionContextsCustomData;
};

sas.stress.wkbench.widget.RiskTable.prototype.isFiltered = function(){
    var binding = this.getBinding("rows");
    var bindingInfo = this.getBindingInfo("rows");
    var modelName = bindingInfo.model;
    var tableDataBindPath = binding.getPath();
    var tableModel = this.getModel(modelName);
    return binding.getLength() !== tableModel.getProperty(tableDataBindPath).length;
};

sas.stress.wkbench.widget.RiskTable.prototype.getPersistenceKey = function(){
    return this.getPersistenceId();
};

sas.stress.wkbench.widget.RiskTable.prototype.clearPersistedRows = function(){
    sessionStorage.removeItem(this.getPersistenceKey());
};

sas.stress.wkbench.widget.RiskTable.prototype.removePersistedRow = function(rowObject, persistedRowsArray){
    var objToRemove = {};
    var matchingAttrs = this.getRowUniqueIdentifiers();
    _.each(matchingAttrs, function(attr){
        objToRemove[attr] = rowObject[attr];
    });
    for (var i = persistedRowsArray.length - 1; i >= 0; i--) {
        if(_.isEqual(persistedRowsArray[i], objToRemove)){
            persistedRowsArray.splice(i,1);
            break; //Done removing, don't loop any more
        }
    }
};

sas.stress.wkbench.widget.RiskTable.prototype.removePersistedRows = function(pathsArray, persistedRowsArray){
    if(_.isEmpty(pathsArray)){
        return persistedRowsArray;
    }
    var that = this;
    var bindingInfo = this.getBindingInfo("rows");
    var modelName = bindingInfo.model;
    var tableModel = this.getModel(modelName);
    var rowObject;
    _.each(pathsArray, function(contextPath){
        rowObject = tableModel.getProperty(contextPath);
        that.removePersistedRow(rowObject, persistedRowsArray);
    });
    return persistedRowsArray;
};

//Remove the entries(related to the paths) from the persisted rows list in sessionStorage
//returns the current selections count
sas.stress.wkbench.widget.RiskTable.prototype.removeAndPersistSelectedRows = function(pathsArray){
    var rowsToPersist = this.removePersistedRows(pathsArray, this.getPersistedRowsArray());
    this.updatePersistedRowsInStorage(rowsToPersist);
    return rowsToPersist;
};

sas.stress.wkbench.widget.RiskTable.prototype.addPersistedRows = function(pathsArray, persistedRowsArray){
    if(_.isEmpty(pathsArray)){
        return persistedRowsArray;
    }
    var bindingInfo = this.getBindingInfo("rows");
    var modelName = bindingInfo.model;
    var tableModel = this.getModel(modelName);
    var contextPath, rowObject, persistObj, exists;
    for(var i=0; i<pathsArray.length; i++){
        contextPath = pathsArray[i];
        rowObject = tableModel.getProperty(contextPath);
        persistObj = this.getObjectToPersist(rowObject);
        exists = _.find(persistedRowsArray, function(persistedRow){
            return _.isEqual(persistedRow, persistObj);
        });
        if(!exists){
            persistedRowsArray.push(persistObj);
        }
    }
    return persistedRowsArray;
};

//This is called when the table data is updated. Now apply the preserved selections persisted in session storage.
sas.stress.wkbench.widget.RiskTable.prototype.applyPersistedRows = function(){
    var persistedRowsArray = this.getPersistedRowsArray();
    if(!(this.getPersistSelectionsThroughSession() && persistedRowsArray.length>0)) {
        return;
    }
    var matchingAttrs = this.getRowUniqueIdentifiers();
    var rowContexts = this.getBinding("rows").getContexts();
    var rowContext, rowObject;
    //Not using third party libraries for following looping to get better performance with native JS
    for(var i=0; i<rowContexts.length; i++){
        rowContext = rowContexts[i];
        rowObject = rowContext.getObject();
        for (var j = persistedRowsArray.length - 1; j >= 0; j--) {
            var match = true;
            _.each(matchingAttrs, function(attr){
                match = match && (rowObject[attr] === persistedRowsArray[j][attr]);
            });
            if(match){
                this.addSelectionInterval(i, i);
                persistedRowsArray.splice(j, 1);
                break;
            }
        }
        if(persistedRowsArray.length === 0){
            break;
        }
    }
};

//This is to pair the selectedContextPaths list with the persistedRows.
sas.stress.wkbench.widget.RiskTable.prototype.updateSelectedContextPaths = function(contexts){
    var binding = this.getBinding("rows");
    var contexts = contexts || binding.getContexts();
    if(_.isEmpty(contexts)){
        return;
    }

    var selectionContextsCustomData = this.findOrAddSelectionContextsCustomData();
    var selectedContextPaths = []; //Reset the array to be filled in fresh from session storage
    var persistedRowsArray = this.getPersistedRowsArray();
    if(_.isEmpty(persistedRowsArray)){ //Nothing saved away for this table, save the rest of the processing
        return;
    }

    var matchingAttrs = this.getRowUniqueIdentifiers();
    var modelName = this.getBindingInfo("rows").model;
    var tableModel = this.getModel(modelName);
    var bindPath = binding.getPath();
    var tableData = tableModel.getProperty(bindPath);
    var rowObject, rowContextPath;
    //Not using third party libraries for following looping to get better performance with native JS
    for(var i=0; i<tableData.length; i++){
        rowObject = tableData[i];
        for (var j = persistedRowsArray.length - 1; j >= 0; j--) {
            var match = true;
            _.each(matchingAttrs, function(attr){
                match = match && (rowObject[attr] === persistedRowsArray[j][attr]);
            });
            if(match){
                rowContextPath = bindPath.match("/$") ? (bindPath + i) : (bindPath + "/" + i);
                selectedContextPaths.push(rowContextPath);
            }
        }
    }
    selectionContextsCustomData.setValue(selectedContextPaths);
    this.fireEvent("selectionUpdated",{"selectionPaths": selectedContextPaths});
};

sas.stress.wkbench.widget.RiskTable.prototype.getPersistedRowsArray = function(){
    var persistedRows = sessionStorage.getItem(this.getPersistenceKey());
    var persistedRowsArray = JSON.parse(persistedRows) || [];
    return persistedRowsArray;
};

sas.stress.wkbench.widget.RiskTable.prototype.getObjectToPersist = function(rowObject){
    var persistObj = {};
    _.each(this.getRowUniqueIdentifiers(), function(attr){
        persistObj[attr] = rowObject[attr];
    });
    return persistObj;
};

sas.stress.wkbench.widget.RiskTable.prototype.getMatchingPersistedRowObject = function(rowObject){
    var match,
        matchingAttrs = this.getRowUniqueIdentifiers(),
        persistedRowsArray = this.getPersistedRowsArray();
    var matchingObj = _.find(persistedRowsArray, function(persistedRow){
        match = true;
        _.each(matchingAttrs, function(attr){
            match = match && (rowObject[attr] === persistedRow[attr]);
        });
        return match;
    });
    return matchingObj;
};

sas.stress.wkbench.widget.RiskTable.prototype.updatePersistedRowsInStorage = function(persistedRowsArray){
    if(persistedRowsArray.length > 0){
        sessionStorage.setItem(this.getPersistenceKey(), JSON.stringify(persistedRowsArray));
    }else{
        sessionStorage.removeItem(this.getPersistenceKey());
    }
};